@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -138,7 +138,6 @@ unsigned name2addr(SOCKADDR_UNION *addr, char *name, int passive) {
     if(retval)
         addrlist2addr(addr, addr_list);
     str_free(addr_list->addr);
-    str_free(addr_list->session);
     str_free(addr_list);
     return retval;
 }
@@ -154,7 +153,6 @@ unsigned hostport2addr(SOCKADDR_UNION *addr,
     if(num)
         addrlist2addr(addr, addr_list);
     str_free(addr_list->addr);
-    str_free(addr_list->session);
     str_free(addr_list);
     return num;
 }
@@ -192,13 +190,10 @@ unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
             s_log(LOG_ERR, "Unix socket path is too long");
             return 0; /* no results */
         }
-        addr_list->addr=str_realloc(addr_list->addr,
+        addr_list->addr=str_realloc_detached(addr_list->addr,
             (addr_list->num+1)*sizeof(SOCKADDR_UNION));
         addr_list->addr[addr_list->num].un.sun_family=AF_UNIX;
         strcpy(addr_list->addr[addr_list->num].un.sun_path, name);
-        addr_list->session=str_realloc(addr_list->session,
-            (addr_list->num+1)*sizeof(SSL_SESSION *));
-        addr_list->session[addr_list->num]=NULL;
         ++(addr_list->num);
         return 1; /* ok - return the number of new addresses */
     }
@@ -223,9 +218,9 @@ unsigned name2addrlist(SOCKADDR_LIST *addr_list, char *name) {
 
 unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
         char *host_name, char *port_name) {
-    struct addrinfo hints, *res=NULL, *cur;
+    struct addrinfo hints, *res, *cur;
     int err, retry=0;
-    unsigned num=0;
+    unsigned num;
 
     memset(&hints, 0, sizeof hints);
 #if defined(USE_IPv6) || defined(USE_WIN32)
@@ -236,19 +231,20 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
     hints.ai_socktype=SOCK_STREAM;
     hints.ai_protocol=IPPROTO_TCP;
     hints.ai_flags=0;
-    if(addr_list->passive) {
-        hints.ai_family=AF_INET; /* first try IPv4 for passive requests */
+    if(addr_list->passive)
         hints.ai_flags|=AI_PASSIVE;
-    }
 #ifdef AI_ADDRCONFIG
     hints.ai_flags|=AI_ADDRCONFIG;
 #endif
     for(;;) {
+        res=NULL;
         err=getaddrinfo(host_name, port_name, &hints, &res);
-        if(!err)
+        if(!err) /* success */
             break;
-        if(res)
-            freeaddrinfo(res);
+        if(err==EAI_SERVICE) {
+            s_log(LOG_ERR, "Unknown TCP service \"%s\"", port_name);
+            return 0; /* error */
+        }
         if(err==EAI_AGAIN && ++retry<=3) {
             s_log(LOG_DEBUG, "getaddrinfo: EAI_AGAIN received: retrying");
             sleep(1);
@@ -260,19 +256,6 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
             continue; /* retry for unconfigured network interfaces */
         }
 #endif
-#if defined(USE_IPv6) || defined(USE_WIN32)
-        if(hints.ai_family==AF_INET) {
-            hints.ai_family=AF_UNSPEC;
-            continue; /* retry for non-IPv4 addresses */
-        }
-#endif
-        break;
-    }
-    if(err==EAI_SERVICE) {
-        s_log(LOG_ERR, "Unknown TCP service \"%s\"", port_name);
-        return 0; /* error */
-    }
-    if(err) {
         s_log(LOG_ERR, "Error resolving \"%s\": %s",
             host_name ? host_name :
                 (addr_list->passive ? DEFAULT_ANY : DEFAULT_LOOPBACK),
@@ -280,23 +263,24 @@ unsigned hostport2addrlist(SOCKADDR_LIST *addr_list,
         return 0; /* error */
     }
 
-    /* copy the list of addresses */
+    /* find the number of newly resolved addresses */
+    num=0;
     for(cur=res; cur; cur=cur->ai_next) {
         if(cur->ai_addrlen>(int)sizeof(SOCKADDR_UNION)) {
             s_log(LOG_ERR, "INTERNAL ERROR: ai_addrlen value too big");
             freeaddrinfo(res);
             return 0; /* no results */
         }
-        addr_list->addr=str_realloc(addr_list->addr,
-            (addr_list->num+1)*sizeof(SOCKADDR_UNION));
-        memcpy(&addr_list->addr[addr_list->num], cur->ai_addr,
-            (size_t)cur->ai_addrlen);
-        addr_list->session=str_realloc(addr_list->session,
-            (addr_list->num+1)*sizeof(SSL_SESSION *));
-        addr_list->session[addr_list->num]=NULL;
-        ++(addr_list->num);
         ++num;
     }
+
+    /* append the newly resolved addresses to addr_list->addr */
+    addr_list->addr=str_realloc_detached(addr_list->addr,
+        (addr_list->num+num)*sizeof(SOCKADDR_UNION));
+    for(cur=res; cur; cur=cur->ai_next)
+        memcpy(&addr_list->addr[(addr_list->num)++], cur->ai_addr,
+            (size_t)cur->ai_addrlen);
+
     freeaddrinfo(res);
     return num; /* ok - return the number of new addresses */
 }
@@ -312,20 +296,18 @@ void addrlist_clear(SOCKADDR_LIST *addr_list, int passive) {
 NOEXPORT void addrlist_reset(SOCKADDR_LIST *addr_list) {
     addr_list->num=0;
     addr_list->addr=NULL;
-    addr_list->session=NULL;
-    addr_list->rr=0; /* reset the round-robin counter */
+    addr_list->start=0;
     addr_list->parent=addr_list; /* allow a copy to locate its parent */
 }
 
 unsigned addrlist_dup(SOCKADDR_LIST *dst, const SOCKADDR_LIST *src) {
     memcpy(dst, src, sizeof(SOCKADDR_LIST));
     if(src->num) { /* already resolved */
-        dst->addr=str_alloc(src->num*sizeof(SOCKADDR_UNION));
+        dst->addr=str_alloc_detached(src->num*sizeof(SOCKADDR_UNION));
         memcpy(dst->addr, src->addr, src->num*sizeof(SOCKADDR_UNION));
     } else { /* delayed resolver */
         addrlist_resolve(dst);
     }
-    /* we currently don't make a local copy of src->session */
     return dst->num;
 }
 
@@ -336,16 +318,13 @@ unsigned addrlist_resolve(SOCKADDR_LIST *addr_list) {
     addrlist_reset(addr_list);
     for(host=addr_list->names; host; host=host->next)
         num+=name2addrlist(addr_list, host->name);
-    switch(num) {
-    case 0:
-    case 1:
-        addr_list->rr=0;
-        break;
-    default:
+    if(num<2) {
+        addr_list->start=0;
+    } else {
         /* randomize the initial value of round-robin counter */
         /* ignore the error value and the distribution bias */
         RAND_bytes((unsigned char *)&rnd, sizeof rnd);
-        addr_list->rr=rnd%num;
+        addr_list->start=rnd%num;
     }
     return num;
 }
@@ -452,7 +431,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
     /* not numerical: need to call resolver library */
     *res=NULL;
     ai=NULL;
-    CRYPTO_w_lock(stunnel_locks[LOCK_INET]);
+    stunnel_write_lock(&stunnel_locks[LOCK_INET]);
 #ifdef HAVE_GETHOSTBYNAME2
     h=gethostbyname2(node, AF_INET6);
     if(h) /* some IPv6 addresses found */
@@ -470,7 +449,7 @@ NOEXPORT int getaddrinfo(const char *node, const char *service,
 #ifdef HAVE_ENDHOSTENT
     endhostent();
 #endif
-    CRYPTO_w_unlock(stunnel_locks[LOCK_INET]);
+    stunnel_write_unlock(&stunnel_locks[LOCK_INET]);
     if(retval) { /* error: free allocated memory */
         freeaddrinfo(*res);
         *res=NULL;
@@ -605,10 +584,10 @@ int getnameinfo(const struct sockaddr *sa, socklen_t salen,
                 (void *)&((struct sockaddr_in *)sa)->sin_addr,
             host, hostlen);
 #else /* USE_IPv6 */
-        CRYPTO_w_lock(stunnel_locks[LOCK_INET]); /* inet_ntoa is not mt-safe */
+        stunnel_write_lock(&stunnel_locks[LOCK_INET]); /* inet_ntoa is not mt-safe */
         strncpy(host, inet_ntoa(((struct sockaddr_in *)sa)->sin_addr),
             hostlen);
-        CRYPTO_w_unlock(stunnel_locks[LOCK_INET]);
+        stunnel_write_unlock(&stunnel_locks[LOCK_INET]);
         host[hostlen-1]='\0';
 #endif /* USE_IPv6 */
     }
