@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2017 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2018 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -64,7 +64,10 @@ struct sockaddr_un {
 };
 #endif
 
-NOEXPORT int accept_connection(SERVICE_OPTIONS *);
+NOEXPORT int accept_connection(SERVICE_OPTIONS *, unsigned);
+NOEXPORT int exec_connect_start(void);
+NOEXPORT void unbind_port(SERVICE_OPTIONS *, unsigned);
+NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *, int, unsigned);
 #ifdef HAVE_CHROOT
 NOEXPORT int change_root(void);
 #endif
@@ -225,8 +228,10 @@ NOEXPORT void client_status(void) { /* dead children detected */
     int pid, status;
 
 #ifdef HAVE_WAIT_FOR_PID
+    s_log(LOG_DEBUG, "Retrieving pid values with " HAVE_WAIT_FOR_PID);
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
 #else
+    s_log(LOG_DEBUG, "Retrieving a pid value with wait()");
     if((pid=wait(&status))>0) {
 #endif
 #ifdef WIFSIGNALED
@@ -254,8 +259,10 @@ void child_status(void) { /* dead libwrap or 'exec' process detected */
     int pid, status;
 
 #ifdef HAVE_WAIT_FOR_PID
+    s_log(LOG_DEBUG, "Retrieving pid values with " HAVE_WAIT_FOR_PID);
     while((pid=wait_for_pid(-1, &status, WNOHANG))>0) {
 #else
+    s_log(LOG_DEBUG, "Retrieving a pid value with wait()");
     if((pid=wait(&status))>0) {
 #endif
 #ifdef WIFSIGNALED
@@ -282,8 +289,14 @@ void child_status(void) { /* dead libwrap or 'exec' process detected */
 /**************************************** main loop accepting connections */
 
 void daemon_loop(void) {
-    if(cron_init()) /* initialize periodic events */
-        fatal("Cron initialization failed");
+    if(cron_init()) { /* initialize periodic events */
+        s_log(LOG_CRIT, "Cron initialization failed");
+        exit(1);
+    }
+    if(exec_connect_start()) {
+        s_log(LOG_CRIT, "Failed to start exec+connect services");
+        exit(1);
+    }
     while(1) {
         int temporary_lack_of_resources=0;
         int num=s_poll_wait(fds, -1, -1);
@@ -295,10 +308,16 @@ void daemon_loop(void) {
             if(s_poll_canread(fds, signal_pipe[0]))
                 if(signal_pipe_dispatch()) /* SIGNAL_TERMINATE or error */
                     break; /* terminate daemon_loop */
-            for(opt=service_options.next; opt; opt=opt->next)
-                if(opt->option.accept && s_poll_canread(fds, opt->fd))
-                    if(accept_connection(opt))
+            for(opt=service_options.next; opt; opt=opt->next) {
+                unsigned i;
+                for(i=0; i<opt->local_addr.num; ++i) {
+                    SOCKET fd=opt->local_fd[i];
+                    if(fd!=INVALID_SOCKET &&
+                            s_poll_canread(fds, fd) &&
+                            accept_connection(opt, i))
                         temporary_lack_of_resources=1;
+                }
+            }
         } else {
             log_error(LOG_NOTICE, get_last_socket_error(),
                 "daemon_loop: s_poll_wait");
@@ -313,15 +332,15 @@ void daemon_loop(void) {
 }
 
     /* return 1 when a short delay is needed before another try */
-NOEXPORT int accept_connection(SERVICE_OPTIONS *opt) {
+NOEXPORT int accept_connection(SERVICE_OPTIONS *opt, unsigned i) {
     SOCKADDR_UNION addr;
     char *from_address;
-    SOCKET s;
+    SOCKET s, fd=opt->local_fd[i];
     socklen_t addrlen;
 
     addrlen=sizeof addr;
     for(;;) {
-        s=s_accept(opt->fd, &addr.sa, &addrlen, 1, "local socket");
+        s=s_accept(fd, &addr.sa, &addrlen, 1, "local socket");
         if(s!=INVALID_SOCKET) /* success! */
             break;
         switch(get_last_socket_error()) {
@@ -356,10 +375,11 @@ NOEXPORT int accept_connection(SERVICE_OPTIONS *opt) {
         return 0;
     }
 #endif
-    if(create_client(opt->fd, s,
-            alloc_client_session(opt, s, s), client_thread)) {
+    service_up_ref(opt);
+    if(create_client(fd, s, alloc_client_session(opt, s, s))) {
         s_log(LOG_ERR, "Connection rejected: create_client failed");
         closesocket(s);
+        service_free(opt);
         return 0;
     }
     return 0;
@@ -367,43 +387,48 @@ NOEXPORT int accept_connection(SERVICE_OPTIONS *opt) {
 
 /**************************************** initialization helpers */
 
+NOEXPORT int exec_connect_start(void) {
+    SERVICE_OPTIONS *opt;
+
+    for(opt=service_options.next; opt; opt=opt->next) {
+        if(opt->exec_name && opt->connect_addr.names) {
+            s_log(LOG_DEBUG, "Starting exec+connect service [%s]",
+                opt->servname);
+            service_up_ref(opt);
+            if(create_client(INVALID_SOCKET, INVALID_SOCKET,
+                    alloc_client_session(opt, INVALID_SOCKET, INVALID_SOCKET))) {
+                s_log(LOG_ERR, "Failed to start exec+connect service [%s]",
+                    opt->servname);
+                service_free(opt);
+                return 1; /* fatal error */
+            }
+        }
+    }
+    return 0; /* OK */
+}
+
 /* clear fds, close old ports */
 void unbind_ports(void) {
     SERVICE_OPTIONS *opt;
-#ifdef HAVE_STRUCT_SOCKADDR_UN
-    struct stat sb; /* buffer for lstat() */
-#endif
 
     s_poll_init(fds);
     s_poll_add(fds, signal_pipe[0], 1, 0);
 
-    for(opt=service_options.next; opt; opt=opt->next) {
-        s_log(LOG_DEBUG, "Closing service [%s]", opt->servname);
-        if(opt->option.accept && opt->fd!=INVALID_SOCKET) {
-            if(opt->fd<(SOCKET)listen_fds_start ||
-                    opt->fd>=(SOCKET)(listen_fds_start+systemd_fds))
-                closesocket(opt->fd);
-            s_log(LOG_DEBUG, "Service [%s] closed (FD=%ld)",
-                opt->servname, (long)opt->fd);
-            opt->fd=INVALID_SOCKET;
-#ifdef HAVE_STRUCT_SOCKADDR_UN
-            if(opt->local_addr.sa.sa_family==AF_UNIX) {
-                if(lstat(opt->local_addr.un.sun_path, &sb))
-                    sockerror(opt->local_addr.un.sun_path);
-                else if(!S_ISSOCK(sb.st_mode))
-                    s_log(LOG_ERR, "Not a socket: %s",
-                        opt->local_addr.un.sun_path);
-                else if(unlink(opt->local_addr.un.sun_path))
-                    sockerror(opt->local_addr.un.sun_path);
-                else
-                    s_log(LOG_DEBUG, "Socket removed: %s",
-                        opt->local_addr.un.sun_path);
-            }
-#endif
-        } else if(opt->exec_name && opt->connect_addr.names) {
+    opt=service_options.next;
+    service_options.next=NULL;
+    service_free(&service_options);
+
+    while(opt) {
+        unsigned i;
+        s_log(LOG_DEBUG, "Unbinding service [%s]", opt->servname);
+        for(i=0; i<opt->local_addr.num; ++i)
+            unbind_port(opt, i);
+        /* exec+connect service */
+        if(opt->exec_name && opt->connect_addr.names) {
             /* create exec+connect services             */
             /* FIXME: this is just a crude workaround   */
             /*        is it better to kill the service? */
+            /* FIXME: this won't work with FORK threads */
             opt->option.retry=0;
         }
         /* purge session cache of the old SSL_CTX object */
@@ -413,17 +438,53 @@ void unbind_ports(void) {
             SSL_CTX_flush_sessions(opt->ctx,
                 (long)time(NULL)+opt->session_timeout+1);
         s_log(LOG_DEBUG, "Service [%s] closed", opt->servname);
+
+        {
+            SERVICE_OPTIONS *garbage=opt;
+            opt=opt->next;
+            garbage->next=NULL;
+            service_free(garbage);
+        }
     }
 }
 
+NOEXPORT void unbind_port(SERVICE_OPTIONS *opt, unsigned i) {
+    SOCKET fd=opt->local_fd[i];
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    SOCKADDR_UNION *addr=opt->local_addr.addr+i;
+    struct stat sb; /* buffer for lstat() */
+#endif
+
+    if(fd==INVALID_SOCKET)
+        return;
+    opt->local_fd[i]=INVALID_SOCKET;
+
+    if(fd<(SOCKET)listen_fds_start ||
+            fd>=(SOCKET)(listen_fds_start+systemd_fds))
+        closesocket(fd);
+    s_log(LOG_DEBUG, "Service [%s] closed (FD=%ld)",
+        opt->servname, (long)fd);
+
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    if(addr->sa.sa_family==AF_UNIX) {
+        if(lstat(addr->un.sun_path, &sb))
+            sockerror(addr->un.sun_path);
+        else if(!S_ISSOCK(sb.st_mode))
+            s_log(LOG_ERR, "Not a socket: %s",
+                addr->un.sun_path);
+        else if(unlink(addr->un.sun_path))
+            sockerror(addr->un.sun_path);
+        else
+            s_log(LOG_DEBUG, "Socket removed: %s",
+                addr->un.sun_path);
+    }
+#endif
+}
+
 /* open new ports, update fds */
 int bind_ports(void) {
     SERVICE_OPTIONS *opt;
-    char *local_address;
     int listening_section;
-#ifdef HAVE_STRUCT_SOCKADDR_UN
-    struct stat sb; /* buffer for lstat() */
-#endif
 
 #ifdef USE_LIBWRAP
     /* execute after options_cmdline() to know service_options.next,
@@ -438,86 +499,41 @@ int bind_ports(void) {
 
     /* allow clean unbind_ports() even though
        bind_ports() was not fully performed */
-    for(opt=service_options.next; opt; opt=opt->next)
-        if(opt->option.accept)
-            opt->fd=INVALID_SOCKET;
+    for(opt=service_options.next; opt; opt=opt->next) {
+        unsigned i;
+        for(i=0; i<opt->local_addr.num; ++i)
+            opt->local_fd[i]=INVALID_SOCKET;
+    }
 
     listening_section=0;
     for(opt=service_options.next; opt; opt=opt->next) {
-        if(opt->option.accept) {
-            if(listening_section<systemd_fds) {
-                opt->fd=(SOCKET)(listen_fds_start+listening_section);
-                s_log(LOG_DEBUG,
-                    "Listening file descriptor received from systemd (FD=%ld)",
-                    (long)opt->fd);
-            } else {
-                opt->fd=s_socket(opt->local_addr.sa.sa_family,
-                    SOCK_STREAM, 0, 1, "accept socket");
-                if(opt->fd==INVALID_SOCKET)
-                    return 1;
-                s_log(LOG_DEBUG, "Listening file descriptor created (FD=%ld)",
-                    (long)opt->fd);
+        opt->bound_ports=0;
+        if(opt->local_addr.num) { /* ports to bind for this service */
+            unsigned i;
+            s_log(LOG_DEBUG, "Binding service [%s]", opt->servname);
+            for(i=0; i<opt->local_addr.num; ++i) {
+                SOCKET fd;
+                fd=bind_port(opt, listening_section, i);
+                opt->local_fd[i]=fd;
+                if(fd!=INVALID_SOCKET) {
+                    s_poll_add(fds, fd, 1, 0);
+                    ++opt->bound_ports;
+                }
             }
-            if(set_socket_options(opt->fd, 0)<0) {
-                closesocket(opt->fd);
-                opt->fd=INVALID_SOCKET;
+            if(!opt->bound_ports) {
+                s_log(LOG_ERR, "Could not bind any accepting port");
                 return 1;
             }
-            /* local socket can't be unnamed */
-            local_address=s_ntop(&opt->local_addr, addr_len(&opt->local_addr));
-            /* we don't bind or listen on a socket inherited from systemd */
-            if(listening_section>=systemd_fds) {
-                if(bind(opt->fd, &opt->local_addr.sa, addr_len(&opt->local_addr))) {
-                    sockerror("bind");
-                    s_log(LOG_ERR, "Error binding service [%s] to %s",
-                        opt->servname, local_address);
-                    closesocket(opt->fd);
-                    opt->fd=INVALID_SOCKET;
-                    str_free(local_address);
-                    return 1;
-                }
-                if(listen(opt->fd, SOMAXCONN)) {
-                    sockerror("listen");
-                    closesocket(opt->fd);
-                    opt->fd=INVALID_SOCKET;
-                    str_free(local_address);
-                    return 1;
-                }
-            }
-#ifdef HAVE_STRUCT_SOCKADDR_UN
-            /* chown the UNIX socket, errors are ignored */
-            if(opt->local_addr.sa.sa_family==AF_UNIX &&
-                    (opt->uid || opt->gid)) {
-                /* fchown() does *not* work on UNIX sockets */
-                if(!lchown(opt->local_addr.un.sun_path, opt->uid, opt->gid))
-                    s_log(LOG_DEBUG,
-                        "Socket chown succeeded: %s, UID=%u, GID=%u",
-                        opt->local_addr.un.sun_path,
-                        (unsigned)opt->uid, (unsigned)opt->gid);
-                else if(lstat(opt->local_addr.un.sun_path, &sb))
-                    sockerror(opt->local_addr.un.sun_path);
-                else if(sb.st_uid==opt->uid && sb.st_gid==opt->gid)
-                    s_log(LOG_DEBUG,
-                        "Socket chown unneeded: %s, UID=%u, GID=%u",
-                        opt->local_addr.un.sun_path,
-                        (unsigned)opt->uid, (unsigned)opt->gid);
-                else
-                    s_log(LOG_ERR, "Socket chown failed: %s, UID=%u, GID=%u",
-                        opt->local_addr.un.sun_path,
-                        (unsigned)opt->uid, (unsigned)opt->gid);
-            }
-#endif
-            s_poll_add(fds, opt->fd, 1, 0);
-            s_log(LOG_DEBUG, "Service [%s] (FD=%ld) bound to %s",
-                opt->servname, (long)opt->fd, local_address);
-            str_free(local_address);
             ++listening_section;
         } else if(opt->exec_name && opt->connect_addr.names) {
-            /* create exec+connect services */
-            /* FIXME: needs to be delayed on reload with opt->option.retry set */
-            create_client(INVALID_SOCKET, INVALID_SOCKET,
-                alloc_client_session(opt, INVALID_SOCKET, INVALID_SOCKET),
-                client_thread);
+            s_log(LOG_DEBUG, "Skipped exec+connect service [%s]", opt->servname);
+#ifndef OPENSSL_NO_TLSEXT
+        } else if(!opt->option.client && opt->sni) {
+            s_log(LOG_DEBUG, "Skipped SNI slave service [%s]", opt->servname);
+#endif
+        } else { /* each service must define two endpoints */
+            s_log(LOG_ERR, "Invalid service [%s]", opt->servname);
+            return 1;
         }
     }
     if(listening_section<systemd_fds) {
@@ -529,6 +545,88 @@ int bind_ports(void) {
     return 0; /* OK */
 }
 
+NOEXPORT SOCKET bind_port(SERVICE_OPTIONS *opt, int listening_section, unsigned i) {
+    SOCKET fd;
+    SOCKADDR_UNION *addr=opt->local_addr.addr+i;
+    char *local_address;
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    struct stat sb; /* buffer for lstat() */
+#endif
+
+    if(listening_section<systemd_fds) {
+        fd=(SOCKET)(listen_fds_start+listening_section);
+        s_log(LOG_DEBUG,
+            "Listening file descriptor received from systemd (FD=%ld)",
+            (long)fd);
+    } else {
+        fd=s_socket(addr->sa.sa_family, SOCK_STREAM, 0, 1, "accept socket");
+        if(fd==INVALID_SOCKET)
+            return INVALID_SOCKET;
+        s_log(LOG_DEBUG, "Listening file descriptor created (FD=%ld)",
+            (long)fd);
+    }
+
+    if(socket_options_set(opt, fd, 0)<0) {
+        closesocket(fd);
+        return INVALID_SOCKET;
+    }
+
+    /* local socket can't be unnamed */
+    local_address=s_ntop(addr, addr_len(addr));
+    /* we don't bind or listen on a socket inherited from systemd */
+    if(listening_section>=systemd_fds) {
+        if(bind(fd, &addr->sa, addr_len(addr))) {
+            if(opt->bound_ports) {
+                log_error(LOG_DEBUG, get_last_socket_error(), "bind");
+                s_log(LOG_DEBUG, "Ignoring error binding service [%s] to %s",
+                    opt->servname, local_address);
+            } else {
+                sockerror("bind");
+                s_log(LOG_ERR, "Error binding service [%s] to %s",
+                    opt->servname, local_address);
+            }
+            str_free(local_address);
+            closesocket(fd);
+            return INVALID_SOCKET;
+        }
+        if(listen(fd, SOMAXCONN)) {
+            sockerror("listen");
+            str_free(local_address);
+            closesocket(fd);
+            return INVALID_SOCKET;
+        }
+    }
+
+#ifdef HAVE_STRUCT_SOCKADDR_UN
+    /* chown the UNIX socket, errors are ignored */
+    if(addr->sa.sa_family==AF_UNIX &&
+            (opt->uid || opt->gid)) {
+        /* fchown() does *not* work on UNIX sockets */
+        if(!lchown(addr->un.sun_path, opt->uid, opt->gid))
+            s_log(LOG_DEBUG,
+                "Socket chown succeeded: %s, UID=%u, GID=%u",
+                addr->un.sun_path,
+                (unsigned)opt->uid, (unsigned)opt->gid);
+        else if(lstat(addr->un.sun_path, &sb))
+            sockerror(addr->un.sun_path);
+        else if(sb.st_uid==opt->uid && sb.st_gid==opt->gid)
+            s_log(LOG_DEBUG,
+                "Socket chown unneeded: %s, UID=%u, GID=%u",
+                addr->un.sun_path,
+                (unsigned)opt->uid, (unsigned)opt->gid);
+        else
+            s_log(LOG_ERR, "Socket chown failed: %s, UID=%u, GID=%u",
+                addr->un.sun_path,
+                (unsigned)opt->uid, (unsigned)opt->gid);
+    }
+#endif
+
+    s_log(LOG_INFO, "Service [%s] (FD=%ld) bound to %s",
+        opt->servname, (long)fd, local_address);
+    str_free(local_address);
+    return fd;
+}
+
 #ifdef HAVE_CHROOT
 NOEXPORT int change_root(void) {
     if(!global_options.chroot_dir)
@@ -591,93 +689,91 @@ NOEXPORT int signal_pipe_init(void) {
 #pragma GCC diagnostic push
 #pragma GCC diagnostic ignored "-Wunused-result"
 #endif /* __GNUC__ */
-void signal_post(int sig) {
+void signal_post(uint8_t sig) {
     /* no meaningful way here to handle the result */
-    writesocket(signal_pipe[1], (char *)&sig, sizeof sig);
+    writesocket(signal_pipe[1], (char *)&sig, 1);
 }
 #ifdef __GNUC__
 #pragma GCC diagnostic pop
 #endif /* __GNUC__ */
 
+/* make a single attempt to dispatch a signal from the signal pipe */
+/* return 1 on SIGNAL_TERMINATE or a fatal error, 0 otherwise */
 NOEXPORT int signal_pipe_dispatch(void) {
-    static int sig;
-    static size_t ptr=0;
+    uint8_t sig=0xff;
     ssize_t num;
     char *sig_name;
 
-    s_log(LOG_DEBUG, "Dispatching signals from the signal pipe");
-    for(;;) {
-        num=readsocket(signal_pipe[0], (char *)&sig+ptr, sizeof sig-ptr);
-        if(num==-1 && get_last_socket_error()==S_EWOULDBLOCK) {
-            s_log(LOG_DEBUG, "Signal pipe is empty");
-            return 0;
-        }
-        if(num==-1 || num==0) {
-            if(num)
-                sockerror("signal pipe read");
-            else
-                s_log(LOG_ERR, "Signal pipe closed");
-            s_poll_remove(fds, signal_pipe[0]);
-            closesocket(signal_pipe[0]);
-            closesocket(signal_pipe[1]);
-            if(signal_pipe_init()) {
-                s_log(LOG_ERR,
-                    "Signal pipe reinitialization failed; terminating");
-                return 1;
+    s_log(LOG_DEBUG, "Dispatching a signal from the signal pipe");
+    num=readsocket(signal_pipe[0], (char *)&sig, 1);
+    if(num!=1) {
+        if(num) {
+            if(get_last_socket_error()==S_EWOULDBLOCK) {
+                s_log(LOG_DEBUG, "Signal pipe is empty");
+                return 0;
             }
-            s_poll_add(fds, signal_pipe[0], 1, 0);
-            s_log(LOG_ERR, "Signal pipe reinitialized");
-            return 0;
+            sockerror("signal pipe read");
+        } else {
+            s_log(LOG_ERR, "Signal pipe closed");
         }
-        ptr+=(size_t)num;
-        if(ptr<sizeof sig) {
-            s_log(LOG_DEBUG, "Incomplete signal pipe read (ptr=%ld)",
-                (long)ptr);
-            return 0;
+        s_poll_remove(fds, signal_pipe[0]);
+        closesocket(signal_pipe[0]);
+        closesocket(signal_pipe[1]);
+        if(signal_pipe_init()) {
+            s_log(LOG_ERR,
+                "Signal pipe reinitialization failed; terminating");
+            return 1;
         }
-        ptr=0;
-        switch(sig) {
+        s_poll_add(fds, signal_pipe[0], 1, 0);
+        s_log(LOG_ERR, "Signal pipe reinitialized");
+        return 0;
+    }
+
+    switch(sig) {
 #ifndef USE_WIN32
-        case SIGCHLD:
-            s_log(LOG_DEBUG, "Processing SIGCHLD");
+    case SIGCHLD:
+        s_log(LOG_DEBUG, "Processing SIGCHLD");
 #ifdef USE_FORK
-            client_status(); /* report status of client process */
+        client_status(); /* report status of client process */
 #else /* USE_UCONTEXT || USE_PTHREAD */
-            child_status();  /* report status of libwrap or 'exec' process */
+        child_status();  /* report status of libwrap or 'exec' process */
 #endif /* defined USE_FORK */
-            break;
+        return 0;
 #endif /* !defind USE_WIN32 */
-        case SIGNAL_RELOAD_CONFIG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
-            if(options_parse(CONF_RELOAD)) {
-                s_log(LOG_ERR, "Failed to reload the configuration file");
-            } else {
-                unbind_ports();
-                log_close();
-                options_apply();
-                log_open();
-                ui_config_reloaded();
-                if(bind_ports()) {
-                    /* FIXME: handle the error */
-                }
-            }
-            break;
-        case SIGNAL_REOPEN_LOG:
-            s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
+    case SIGNAL_RELOAD_CONFIG:
+        s_log(LOG_DEBUG, "Processing SIGNAL_RELOAD_CONFIG");
+        if(options_parse(CONF_RELOAD)) {
+            s_log(LOG_ERR, "Failed to reload the configuration file");
+        } else {
+            unbind_ports();
             log_close();
+            options_free(); /* FIXME: the pattern should be copy-apply-free */
+            options_apply();
             log_open();
-            s_log(LOG_NOTICE, "Log file reopened");
-            break;
-        case SIGNAL_TERMINATE:
-            s_log(LOG_DEBUG, "Processing SIGNAL_TERMINATE");
-            s_log(LOG_NOTICE, "Terminated");
-            return 1;
-        default:
-            sig_name=signal_name(sig);
-            s_log(LOG_ERR, "Received %s; terminating", sig_name);
-            str_free(sig_name);
-            return 1;
+            ui_config_reloaded();
+            if(bind_ports()) {
+                /* FIXME: handle the error */
+            }
+            if(exec_connect_start()) {
+                /* FIXME: handle the error */
+            }
         }
+        return 0;
+    case SIGNAL_REOPEN_LOG:
+        s_log(LOG_DEBUG, "Processing SIGNAL_REOPEN_LOG");
+        log_close();
+        log_open();
+        s_log(LOG_NOTICE, "Log file reopened");
+        return 0;
+    case SIGNAL_TERMINATE:
+        s_log(LOG_DEBUG, "Processing SIGNAL_TERMINATE");
+        s_log(LOG_NOTICE, "Terminated");
+        return 1;
+    default:
+        sig_name=signal_name(sig);
+        s_log(LOG_ERR, "Received %s; terminating", sig_name);
+        str_free(sig_name);
+        return 1;
     }
 }
 
@@ -801,7 +897,8 @@ void stunnel_info(int level) {
     if(strcmp(OPENSSL_VERSION_TEXT, OpenSSL_version(OPENSSL_VERSION))) {
         s_log(level, "Compiled with " OPENSSL_VERSION_TEXT);
         s_log(level, "Running  with %s", OpenSSL_version(OPENSSL_VERSION));
-        s_log(level, "Update OpenSSL shared libraries or rebuild stunnel");
+        if((OpenSSL_version_num()^OPENSSL_VERSION_NUMBER)&~0xfffu)
+            s_log(level, "Update OpenSSL shared libraries or rebuild stunnel");
     } else {
         s_log(level, "Compiled/running with " OPENSSL_VERSION_TEXT);
     }
